/*
 * Paths of the configuration files.
 * This file may be changed by the user as needed.
 * There are three empty lines between each definition.
 * These ensure that "local" patches and official patches have
 * only a very low probability of conflicting.
 */


#define CONF_FILE "/etc/mtools.conf"


#define OLD_CONF_FILE "/etc/mtools"



#define LOCAL_CONF_FILE "/etc/default/mtools.conf"
/* Use this if you like to keep the configuration file in a non-standard
 * place such as /etc/default, /opt/etc, /usr/etc, /usr/local/etc ...
 */

#define SYS_CONF_FILE SYSCONFDIR "/mtools.conf"

#define OLD_LOCAL_CONF_FILE "/etc/default/mtools"



#define CFG_FILE1 "/.mtoolsrc"



/* END */
